% function [Mi,Mbc,Bbc] = CylPoissonMatrix(x,r,U0)
%
% This function constructs a matrix representation of the discretized
% Poisson equation in cylindrical coordinates. Let p(x,r) denote
% a function of longitudinal and radial distance (x,r).  The
% Poisson equation in cylindrical coordinates is:
%    d^2(p)/dx^2 - (1/r)*(dp/dr) + d^2(p)/dr^2 = f(x,r)
% where f is a given function and the derivatives should be interpreted
% as partials.
%
% Assume the spatial grid consists of (Nx,Nr) points uniformly spaced
% by (dx,dr) in the (x,r) directions.  The discretized Poisson equation
% at a point (xi,rj) in the "interior" of the domain (i.e. 2<=i<=Nx-1 and
% 2<=j<=Nr-1) is given by:
%
% f(i,j) = p(i-1,j)/dx^2 -2*(1/dx^2+1/dr^2)*p(i,j) + p(i+1,j)/dx^2
%      + (1/dr^2+1/2/dr/r(j))*p(i,j-1) + (1/dr^2-1/2/dr/r(j))*p(i,j+1)
%
% The values of p can be stacked into a long column P of length Nx*Nr.
% There are (Nx-2)*(Nr-2) equations for the interior points.  These
% can be stacked together to form an algebraic equation:
% (1)    Mi*P = F
% where F stacks together the values of f(i,j) at the interior points.
% The boundary conditions are used to generate additional equations for
% points on the edge of the domain (i.e. i=1, Nx or j=1,Nr).  There
% will be  2Nx+2Nr-4 equations from the boundary points and these
% equations can be stacked together to form an algebraic equation:
% (2)   Mbc*P = Bbc
% Equations (1) and (2) combine to form Nx*Nr linear equations in
% Nx*Nr unknowns.  Thus (assuming linearly independence) these equations
% can be solved for a unique vector P.
%
% This function returns the matrices Mi, Mbc, and Bbc used in the
% algebraic equations described above.  These are stored as sparse
% matrices.  The specific boundary condition options are commented
% further in the code.
function [Mi,Mbc,Bbc] = CylPoissonMatrix(x,r,U0)

%% Grid Information
Nx = length(x);             % # of points in x direction
dx = x(2)-x(1);             % Grid spacing in x direction, m
xmax = max(x);              % Maximum x value, m

dr = r(2)-r(1);             % Grid spacing in r direction, m
Nr = length(r);             % # of points in r direction
rmax = max(r);              % Maximum r value, m

%% Poisson Matrix
% Form matrix Mi used in discretized Poisson equation for "interior"
% points. The boundary conditions are handled separately below.
% XXX-This code could be vectorized.
Nnz = 5;            % # of nonzero entries per row
row = zeros(Nnz*(Nx-2)*(Nr-2),1);
col = row;
val = row;

% Discretized Poisson: Nonzero entries are specfied as (row,col,val)
cnt = 1;
for j=2:(Nr-1)
    for i=2:(Nx-1)
        idx = (cnt-1)*Nnz + (1:Nnz);
        row(idx) = cnt;
        col(idx) = [(i-1)+(j-1)*Nx, i+(j-1)*Nx, (i+1)+(j-1)*Nx, ...
            i+(j-2)*Nx, i+j*Nx];
        val(idx) = [1/dx^2, -2*(1/dx^2+1/dr^2), 1/dx^2, ...
            1/dr^2+1/2/dr/r(j), 1/dr^2-1/2/dr/r(j)];
        
        cnt = cnt+1;
    end
end

% Create Sparse Poisson Matrix, Mi
rowdim = (Nx-2)*(Nr-2);
coldim = Nx*Nr;
Mi = sparse(row,col,val,rowdim,coldim);



%% Boundary Conditions
% Construct boundary conditions for inflow (i=1), outflow (i=Nx),
% symmetry axis (j=1), and lateral boundary (j=Nr).
% XXX-Again, this code could be vectorized.

row = [];
col = row;
val = row;

% Add Boundary Conditions at symmetry axis (j=1)
if 1
    % A) v = 0 --> d(Psi)/dx  = 0
    %          --> Psi = function(r)
    % B) du/dr = 0 --> d^2(Psi)/dr^2 - (1/r)*d(Psi)/dr = 0
    %              --> d(Psi)/dr = 0 because r=0 on the symmetry axis
    % Note: A) and B) imply that Psi is a constant on the symmetry axis.
    %       --> Psi(i,1) = Psi(1,1) = 0 [See Inflow Boundary Condition]
    Bbc = [];
    cnt = 1;
    for i=1:Nx
        row = [row; cnt];
        col = [col; i];
        val = [val; 1];
        Bbc = [Bbc; 0];
        cnt = cnt+1;
    end
else
    % du/dr = 0 --> d^2(Psi)/dr^2 - (1/r)*d(Psi)/dr = 0
    %           --> d(Psi)/dr = 0 because r=0 on the symmetry axis
    % A one-sided first derivative with second order accuracy is
    % used below.  See Eq 4.13 in Zikanov's textbook.
    % This boundary condition gave odd edge effects at the symmetry axis.
    % Specifically the radial velocity had large variations near r=0.
    Bbc = [];
    cnt = 1;
    for i=1:Nx
        row = [row; cnt; cnt; cnt];
        col = [col; i; i+Nx; i+2*Nx];
        val = [val; -3/2/dr; 4/2/dr; -1/2/dr];
        Bbc = [Bbc; 0];
        cnt = cnt+1;
    end
end

% Add Boundary Conditions at lateral boundary (j=Nr)
%     u = U0 --> d(Psi)/dr = -r*U0
% A one-sided first deriv. with second order accuracy is used below
% to enforce d(Psi)/dr=-r*U0.  See Eq 4.13 in Zikanov's textbook.
%
% Note: This implies that Psi=-0.5*r^2*U0 + function(x) where the
% function of x is related to the radial velocity v.
% Sorensen & Myken give the condition d(rv)/dr = 0 on this boundary.
% Thus the value of the radial velocity is not directly constrained on
% this boundary and hence we can't impose a boundary condition
% directly on Psi. Instead we impose a BC on the derivative of Psi.
for i=1:Nx
    row = [row; cnt; cnt; cnt];
    col = [col; i+(Nr-1)*Nx; i+(Nr-2)*Nx; i+(Nr-3)*Nx];
    val = [val; 3/2/dr; -4/2/dr; 1/2/dr];
    Bbc = [Bbc; -rmax*U0];
    cnt = cnt+1;
end

% Add Boundary Conditions at inflow boundary (i=1)
%   A)  u = U0 --> Psi = -0.5*r^2*U0 + function(x)
%   B)  dv/dx = 0 --> d^2(Psi)/dx^2 = 0
% Note 1: A) and B) imply that Psi = -0.5*r^2*U0 + c1*x+c2
% for some constants c1 and c2.  The constant term in Psi has no
% effect since the velocity field only depends on the derivative of
% Psi.  Thus we can set c2=0 without loss of generality.  Moreover, the
% term c1*x is equal to zero because x=0 on the inflow boundary. 
% Thus the boundary condition below enforces Psi = -0.5*r^2*U.
%
% Note 2: Psi = -0.5*r^2*U implies that Psi(1,1) = 0.  This value
% is used in one of the other BC specifications.
for j=2:Nr-1
    row = [row; cnt];
    col = [col; 1+(j-1)*Nx];
    val = [val; 1];
    Bbc = [Bbc; -0.5*r(j).^2*U0];
    cnt = cnt+1;
end

% Add Boundary Conditions at outflow boundary
%     dv/dx = 0 --> d^2(Psi)/dx^2 = 0
% One-sided second derivative with second order accuracy is
% used below.  See Eq 4.20 in Zikanov's textbook.
for j=2:Nr-1
    row = [row; cnt; cnt; cnt; cnt];
    col = [col; Nx+(j-1)*Nx; (Nx-1)+(j-1)*Nx; (Nx-2)+(j-1)*Nx; (Nx-3)+(j-1)*Nx];
    val = [val; 2/dx^2; -5/dx^2; 4/dx^2; -1/dx^2];
    Bbc = [Bbc; 0];
    cnt = cnt+1;
end


% Create Sparse Poisson Matrix for boundary conditions
rowdim = cnt-1;
coldim = Nx*Nr;
Mbc = sparse(row,col,val,rowdim,coldim);
